import collection.Set
//import collection.immutable.Map
//import collection.mutable.Map
import Powerset._
import Color._
import RelationSet._
import Search._

object Graph {
  
  // hypergraph, may have arbitrary edge set
  abstract class Hypergraph [V, E] {
    
    def vertices: Seq[V]
    def edges: Seq[E]
    
  }
  
  // graph, having pairs of vertices as edges
  class Graph [V] () extends Hypergraph [V, (V, V)] {
    
    var vertices : Seq[V] = _
    var edges : Seq[(V,V)] = _

    def this (v: Seq[V], e: Seq[(V,V)]) = {this(); this.vertices = v; this.edges = e} 
    
    def k_vertex_set_solutions (k:Int, p: Seq[V] => Boolean) : Seq[Seq[V]] = {
      (((powerset (vertices)) 
          filter (u => (u.size == k)))
          filter p)
    }
          
    def min_set_solutions ( p: Seq[V] => Boolean) : Seq[Seq[V]] = {
      var m : Seq[Seq[V]] = Seq()
      var k = 0
      do {
        m = (k_vertex_set_solutions(k, p));
        k = k + 1
      }
      while (m.isEmpty && k <= vertices.size)
      return m
    }
    
    def max_set_solutions ( p: Seq[V] => Boolean) : Seq[Seq[V]] = {
      var m : Seq[Seq[V]] = Seq()
      var k = vertices.size
      do {
        m = (k_vertex_set_solutions(k, p));
        k = k - 1
      }
      while (m.isEmpty && k >= 0) 
      return m
    }
    
    // restrict a graph to the members of s
    def restrict (s: Seq[V]) : Graph[V] = {
      new Graph(vertices intersect s, edges filter (e => (s contains e._1) && (s contains e._2)))
    }
    
    def successor (v: V) : Seq[V] = {
      (edges filter (e => (v == e._1))) map (e => e._2)
    }
    
    def degree (v: V) : Int = (successor(v)).size
    
    /***************************************************************/     
    /*****************   SCCs and connectedness ********************/ 
    /***************************************************************/ 
    
    def sccs : Set [Set [V]] = 
      quotient (vertices.toSet) (restrict_equivrel (refl_transclos_idrel (edges.toSet, vertices.toSet)))

    
    /***************************************************************/     
    /*****************        COLORING *****************************/ 
    /***************************************************************/ 
    
    // TODO: note that the order of elements in the vertex list is important.
    // In this case, it is not even important to carry along the order
    // removed this argument:(less : V => V => Boolean)
    
    // Coloring UGraph(Seq(5,6,1,2,4,3), Seq((1,2),(1,6),(3,4),(3,6),(5,2),(5,4),(5,6)))
    // requires 4 colors
    // Coloring UGraph(Seq(1,2,3,4,5,6), Seq((1,2),(1,6),(3,4),(3,6),(5,2),(5,4),(5,6)))
    // requires 2 colors
    def greedyColor : Map [V,Int] = {
      if (vertices.isEmpty)
        Map()
      else {
        var u = vertices.head;
        var s = vertices.tail;
        var c = restrict(s).greedyColor;
        println("u: " + u)
        println("vertices: " + restrict(s).vertices);
        println("edges: " + restrict(s).edges);
        println("c: " + c);
        println("s: " + s);
        println("succ: " + successor(u));
        var fk = c filterKeys (v => (successor(u) intersect s) contains v) ;
        println("filterkeys: " + fk);
        println("minunused: " + min_unused(fk));
        println("------------");
        c + (vertices.head -> 
            min_unused(c filterKeys (v => (successor(u) intersect s) contains v)))
      } 
    }

  }

  
  // undirected graph
  // class UGraph [V] (v: Seq[V], e: Seq[(V,V)]) extends Graph [V] 
  class UGraph [V] () extends Graph [V]	()  	{
    
    // secondary constructor
    def this (v: Seq[V], e: Seq[(V,V)]) = {this(); this.vertices = v; this.edges = (e ++ (e map (p => (p._2, p._1))))} 

    def complement () : UGraph[V] = {
      new UGraph(vertices, 
          vertices.flatMap 
          (u => vertices.flatMap 
              (v => if ((u == v) || (edges contains (u,v))) Seq() else Seq((u,v)))))
    }
    
    // restrict a graph to the members of s
    // TODO: how to inherit that correctly ??
    def ug_restrict (s: Seq[V]) : UGraph[V] = {
      new UGraph(vertices intersect s, edges filter (e => (s contains e._1) && (s contains e._2)))
    }
    
    
    //override var vertices = v
    // TODO: one would like to have removeDuplicates
    //override var edges = (e ++ (e map (p => (p._2, p._1))))
    
	def is_clique(s: Seq[V]) : Boolean = 
	  (s forall (u =>
	    s forall (v =>
	      	(u == v) || (edges contains (u,v)))))
	      	
	 def comp_cliques() : Seq[Seq[V]] = 
	   powerset(this.vertices) filter is_clique
	   
	   
	 // extend a clique cl by a node v
	 def extend_clique(cl: Seq[V], v:V) : Seq[Seq[V]] =
	   if (is_clique(v +: cl))
	     Seq(v +: cl)
	   else
	     Seq(cl,(v +: (cl filter (u => edges contains (u, v)))))
	     
	 def contains_subset (ss: Seq[Seq[V]], s: Seq[V]) : Boolean =
	   ss exists (sup => (s.toSet) subsetOf (sup.toSet))
	
	 def remove_redundant (sl: Seq[Seq[V]], sr: Seq[Seq[V]]) : Seq[Seq[V]] =
	   if (sr.isEmpty)
	     sl
	   else {
	     if ( contains_subset(sl, sr.head) || contains_subset(sr.drop(1),sr.head) )
	       remove_redundant (sl, sr.drop(1))
	     else 
	       remove_redundant (sr.head +: sl, sr.drop(1))
	   }
	       
	   
	 def max_cliques_aux (s: Seq[V]) : Seq[Seq[V]] = 
	   if (s.isEmpty)
	     Seq(Seq())
	   else {
	     val subcliques = max_cliques_aux (s.drop(1));
	     println(subcliques);
	     remove_redundant(Seq(), subcliques.flatMap (cl => extend_clique(cl, s.head)))
	     //subcliques.flatMap (cl => extend_clique(cl, s.head))
	   }
	      	
	 // sequence of all maximal cliques of graph
	 def max_cliques : Seq[Seq[V]] = max_cliques_aux (vertices)
	 
	 // sequence of all maximal independent sets of graph
	 def max_independents : Seq[Seq[V]] = complement.max_cliques
	 
	 def minimal [A] (s: Seq[A], f: A => Int) : Int = 
	   if (s.drop(1).isEmpty)
	     f(s.head)
	   else {
	     val m = minimal(s.drop(1), f);
	     val fh = f(s.head);
	     if (fh < m)
	       fh
	     else
	       m
	   }
	 
	 // TODO: generalize minimal and minimal_val
	 def minimal_val [A,B] (s: Seq[A], f: A => (Int,B)) : ((Int, B), A) = 
	   if (s.drop(1).isEmpty)
	     (f(s.head), s.head)
	   else {
	     val m = minimal_val(s.drop(1), f);
	     val fh = f(s.head);
	     if ((fh._1) < (m._1)._1)
	       (fh, s.head)
	     else
	       m
	   }
	 
	 def chi : Int =  
	   if (vertices.isEmpty)
	     0
	   else 
	     1 + minimal(max_independents, 
	         ((w : Seq[V]) => ug_restrict(vertices diff w).chi))
	   
	 def min_coloring : (Int, Map[Seq[V], Int]) =  
	   if (vertices.isEmpty)
	     (0, Map())
	   else {
	     val r = minimal_val(max_independents, ((w : Seq[V]) => ug_restrict(vertices diff w).min_coloring));
	     val new_c = 1 + r._1._1;
	     val r_map = r._1._2;
	     val sel_w = r._2;
	         
	     (new_c, r_map + (sel_w -> new_c)) 
	   }
	     
  }
  
  // directed graph with adjacency list representation
  // The constructor requires a key-value pair for each vertex, 
  // also for vertices with no successors 
  class AGraph [V] (ve: Map[V, List[V]]) extends Graph [V] {
    
    this.vertices = ve.keys.toSeq
    this.edges = ve.toList.flatMap(kv => kv._2 map (x => (kv._1,x)))
    
    // vertex-edge-map
    var ve_map = ve
    
    // Depth-first search for a path in a graph, from node "from" to node "to"  
    // The function produces a stream of paths. 
    // Each path contains the nodes in reverse order
    def search_path (from: V, to: V): Stream[List[V]] = {
    
    	def search_path_next (xs: List[V]) : List[List[V]] =
    			ve(xs.head).map(n => n::xs)
  
    	def search_sol (v:V) (xs: List[V]) : Boolean = (xs.head == v)

      depthfirst (search_path_next) (search_sol (to)) (List(from))
    }

    
    def dfs_traversal : (List[(V,V)], Map[V,Int], Map[V, Int]) = {
      
      var dfsnum_count = 0
      var compnum_count = 0
      //var dfsnum : collection.mutable.Map[V,Int] = collection.mutable.Map.empty
      //var compnum : collection.mutable.Map[V,Int] = collection.mutable.Map.empty
      var dfsnum : Map[V,Int] = Map.empty
      var compnum : Map[V,Int] = Map.empty
      var tree : List[(V,V)] = List()
      
      def dfs_trav_nodes (v: V) {
        
        dfsnum_count = dfsnum_count + 1;
        dfsnum += (v -> dfsnum_count)
        
        for (w <- ve_map(v)) {
          (dfsnum get w) match {
          case None => {tree = (v, w) :: tree; dfs_trav_nodes(w)}
          case Some(_) => /* skip */
          }
        }
        
        compnum_count = compnum_count + 1;
        compnum += (v -> compnum_count)        
      }
      
      for (v <- vertices)
        if ((dfsnum get v) == None) 
          dfs_trav_nodes(v)  
           
      return (tree, dfsnum, compnum)
    }
    
  }
}
