import collection.Set


object RelationSet {

  
  def bigUnion [A,B] (f: A => Set[B]) (xs: Set[A]) : Set [B] = 
    xs.foldLeft(Set[B]()) ((x,y) => x union (f(y)))
    
  def set_elem_comp [A,B,C] (xs: Set[(A,B)]) (y: (B,C)) : Set[(A,C)] = 
    bigUnion [(A,B), (A,C)] (x => if (x._2 == y._1) Set((x._1, y._2)) else Set()) (xs)
    
  def elem_set_comp [A,B,C]  (x: (A,B)) (ys: Set[(B,C)]) : Set[(A,C)] = 
    bigUnion [(B,C), (A,C)] (y => if (x._2 == y._1) Set((x._1, y._2)) else Set()) (ys)
    
  def rel_comp [A,B,C] (xs: Set[(A,B)]) (ys: Set[(B,C)]) : Set[(A,C)] = 
    bigUnion ((x:(A,B)) => elem_set_comp [A,B,C] (x)(ys)) (xs)

  def rel_comp2 [A,B,C] (xs: Set[(A,B)]) (ys: Set[(B,C)]) : Set[(A,C)] =
    bigUnion (set_elem_comp [A,B,C] (xs)) (ys)
    
    
  def lfp [A] (f: Set[A] => Set[A], xs: Set[A]) : Set [A] = {
    if (f(xs) subsetOf xs)
      xs
    else lfp (f, f(xs))
  }
  

  def relDomain [A,B] (r: Set[(A,B)]) : Set[A] = 
    bigUnion [(A,B),A] (x => Set(x._1)) (r) 
    
  def relRange [A,B] (r: Set[(A,B)]) : Set[B] = 
    bigUnion [(A,B),B] (x => Set(x._2)) (r) 

  def relField [A] (r: Set[(A,A)]) : Set[A] = 
    (relDomain (r)) union (relRange (r))
    
  def idrel [A] (xs: Set[A]) : Set[(A,A)] =
    xs map (x => (x,x))
  
  def rel_prod [A] (r: Set[(A,A)]) (n:Int) : Set [(A,A)] =
    if (n == 0) 
      idrel(relField(r))
    else
      rel_comp(rel_prod(r)(n-1)) (r)
  
    
  def transclos [A] (r: Set[(A,A)]) : Set[(A,A)] = 
    lfp ((ys: Set[(A,A)]) => (ys union rel_comp(ys)(r)), r)
  
  def reflclos  [A] (r: Set[(A,A)]) : Set[(A,A)] =
    (relField (r)) map (x => (x,x)) 
    
    /*
  def refl_transclos [A] (r: Set[(A,A)]) :  Set[(A,A)] = 
    (reflclos (r)) union (transclos (r))
    */
    
  def refl_transclos [A] (r: Set[(A,A)]) :  Set[(A,A)] =
    lfp ((ys: Set[(A,A)]) => (ys union rel_comp(ys)(r)), idrel(relField(r)))
    
  /* Reflexive-transitive closure of a relation initialized with identity relation over xs
   * Useful for taking into account isolated nodes not occurring in r 
   * (such as computation of SCCS)
   */
  def refl_transclos_idrel [A] (r: Set[(A,A)], xs: Set[A]) :  Set[(A,A)] =
    lfp ((ys: Set[(A,A)]) => (ys union rel_comp(ys)(r)), idrel(xs))
    
  def relImage_elem [A,B] (r: Set[(A,B)]) (y: A) : Set[B] = 
    bigUnion [(A,B), B] (x => if (x._1 == y) Set(x._2) else Set()) (r)

  def relImage [A,B] (r: Set[(A,B)]) (xs: Set[A]) : Set[B] = 
    bigUnion (relImage_elem (r)) (xs)

  def reach [A] (r: Set[(A,A)]) (s: Set[A]) : Set[A] =
    lfp ((a: Set[A]) => a union relImage(r)(a), s)
    
    
  def quotient [A] (xs: Set[A]) (r: Set[(A,A)]) : Set [Set[A]] =
    bigUnion [A, Set[A]] (x => Set(relImage_elem(r)(x))) (xs)

  def restrict_equivrel [A] (r: Set[(A,A)]) : Set[(A,A)] =
    r filter (x => r contains (x._2, x._1))
  
  
}